/*
 * 'lander.c' - fly across a fractal terrain...
 *
 * Usage:
 *
 *  lander [-f]
 *
 * Options:
 *
 *  -f = use full screen...
 *
 * Revision History:
 *
 *  Who                 When          What
 *  ------------------- ------------- --------------
 *  Mike Sweet          6/18/91       initial coding
 *  Mike Sweet          6/20/91       added scoring
 *
 */


/*
 * Include necessary headers...
 */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <math.h>
#include <time.h>

#include <gl/gl.h>
#include <gl/device.h>


/*
 * Program constants; self-explainatory...
 */

#define TERRAIN_SIZE    40    /* Size of the terrain map... */
#define TERRAIN_CHANGE  550.0 /* Amount to change height of terrain... */
#define TERRAIN_SPACING 800.0 /* Space between terrain squares... */

#define MAX_THRUST  10.0      /* Maximum x/z thrust of 'ship' */
#define MAX_DELTAH  12.0      /* Maximum vertical thrust... */
#define MAX_DELTAR  90        /* Maximum turning thrust... */

#define GRAVITY     2.0       /* Effective gravity... */
#define INIT_FUEL   2000      /* Initial fuel... */
#define FUEL_MASS   500.0     /* Effect of fuel mass on acceleration... */


/*
 * Globals:
 *
 *  Name            Type                   Description
 *  --------------- ---------------------- ------------------------------
 *  position        Coord [3][3]           current lander position
 *  thrust          Coord                  current forward thrust
 *  rotation        Angle [3]              current lander rotation
 *  deltar          Angle [3]              current change in rotation
 *  deltas          Coord [3]              current x, y, and z motion
 *  view            Angle [3]              current viewing angle
 *  score           long                   current score
 *  fuel            long                   current fuel level
 *  landed          Boolean                TRUE if the user landed
 *  landings        long                   # of landings
 *  terrain         Coord [TERRAIN_SIZE]2  landscape terrain heights
 *  terrain_object  Object                 GL object containing terrain
 *  clear_object    Object                 GL object to erase terrain
 *  clear_count     int                    # of times window needs cleaning
 *  buf_num         int                    current buffer in display
 *  xorg, yorg      long                   current window x, y origin
 *  xsize, ysize    long                   current window x, y size
 *  icon_flag       Boolean                TRUE if the window is iconified
 *  gravity_flag    Boolean                TRUE if gravity is on
 *  current_focus   long                   window with input focus
 *  win             long                   lander window
 *  cursor          unsigned short []      array containing cursor bitmap
 *
 * Notes:
 *
 *  For variables with a dimension of 3, the 0th and 1st indices contain the
 *  previously displayed values for the corresponding buffer, while the 2nd
 *  index contains the current values.
 *
 */

Coord position[3][3];
Coord thrust;
Angle rotation[3];
Angle deltar[3];
Coord deltas[3];
Angle view[3];
long score;
long fuel;
Boolean landed;
long landings;

Coord terrain[TERRAIN_SIZE][TERRAIN_SIZE];
Object terrain_object;
Object clear_object;

int clear_count;
int buf_num;

long xorg, yorg;   /* window origin */
long xsize, ysize; /* window size... */

Boolean icon_flag, gravity_flag;

long current_focus;
long win;


unsigned short cursor[] =
{
 0x0180,
 0x0000,
 0x0000,
 0x0180,
 0x0000,
 0x0000,
 0x0180,
 0xfc3f,
 0xfc3f,
 0x0180,
 0x0000,
 0x0000,
 0x0180,
 0x0000,
 0x0000,
 0x0180
};


/*
 * 'divide_terrain()' - a recursive function which will generate
 *                      the fractal terrain.  Basically, we set
 *                      the corners == 0 and call this function.
 *                      Each generation finds the midpoints of the
 *                      the previous square and moves them up/down...
 */

void divide_terrain(int x1, int y1, int x2, int y2)
{
 int hx, hy;


 hx = (x1 + x2) / 2; /* find the mid-points... */
 hy = (y1 + y2) / 2;

 if (x1 == hx || y1 == hy) /* 'stop' condition.... */
  return;

/*
 Move each midpoint up/down...
*/

 terrain[x1][hy] = (terrain[x1][y1] + terrain[x1][y2]) / 2 +
                   (rand() & 0xffff) * TERRAIN_CHANGE / 65536.0 -
                   TERRAIN_CHANGE * 0.25;
 terrain[x2][hy] = (terrain[x2][y1] + terrain[x2][y2]) / 2 +
                   (rand() & 0xffff) * TERRAIN_CHANGE / 65536.0 -
                   TERRAIN_CHANGE * 0.25;
 terrain[hx][y1] = (terrain[x1][y1] + terrain[x2][y1]) / 2 +
                   (rand() & 0xffff) * TERRAIN_CHANGE / 65536.0 -
                   TERRAIN_CHANGE * 0.25;
 terrain[hx][y2] = (terrain[x1][y2] + terrain[x2][y2]) / 2 +
                   (rand() & 0xffff) * TERRAIN_CHANGE / 65536.0 -
                   TERRAIN_CHANGE * 0.25;
 terrain[hx][hy] = (terrain[hx][y2] + terrain[hx][y1] +
                   terrain[x1][hy] + terrain[x2][hy]) / 4 +
                   (rand() & 0xffff) * TERRAIN_CHANGE / 65536.0 -
                   TERRAIN_CHANGE * 0.25;

/*
 And then divide each quadrant that remains...
*/

 divide_terrain(x1, y1, hx, hy);
 divide_terrain(hx, y1, x2, hy);
 divide_terrain(x1, hy, hx, y2);
 divide_terrain(hx, hy, x2, y2);
}


/*
 * 'reset_window()' - resets the perspective and viewport of the window...
 */

void reset_window(void)
{
 clear_count = 2;

 reshapeviewport();

 getorigin(&xorg, &yorg);
 getsize(&xsize, &ysize);
}


/*
 * 'initialize()' - initializes globals and creates the window for display...
 */

void initialize(void)
{
 int x, y;


/*
 * First, clear the terrain map and generate the fractal terrain...
 */

 srand(time(NULL));

 memset(terrain, 0, sizeof(terrain));
 divide_terrain(0, 0, TERRAIN_SIZE - 1, TERRAIN_SIZE - 1);

 for (x = 0; x < TERRAIN_SIZE; x ++) /* trim the edges... */
  {
   terrain[x][0] =
   terrain[x][TERRAIN_SIZE - 1] =
   terrain[0][x] =
   terrain[TERRAIN_SIZE - 1][x] = 0.0;
  };

 for (x = 0; x < TERRAIN_SIZE; x ++)
  for (y = 0; y < TERRAIN_SIZE; y ++) /* clip lines below sea level... */
   if (terrain[x][y] < 0.0)
    terrain[x][y] = 0.0;

/*
 * Next, initialize the variables controlling the 'ship' position and
 * heading...
 */

 position[0][0] = 0.0;
 position[1][0] = 1000.0;
 position[2][0] = 0.0; /* start at (0, 100, 0)... */

 position[0][1] = 0.0;
 position[1][1] = 1000.0;
 position[2][1] = 0.0; /* start at (0, 100, 0)... */

 position[0][2] = 0.0;
 position[1][2] = 1000.0;
 position[2][2] = 0.0; /* start at (0, 100, 0)... */

 thrust      = 0.0; /* not moving... */

 rotation[0] = 0;   /* facing 'north'... */
 rotation[1] = 0;   /* facing 'north'... */
 rotation[2] = 0;   /* facing 'north'... */

 view[0]     = 0;   /* looking straight ahead... */
 view[1]     = 0;   /* looking straight ahead... */
 view[2]     = 0;   /* looking straight ahead... */

 deltas[0]   = 0.0; /* no x motion... */
 deltas[1]   = 0.0; /* no y motion... */
 deltas[2]   = 0.0; /* no z motion... */

 deltar[0]   = 0;   /* not turning... */
 deltar[1]   = 0;   /* not turning... */
 deltar[2]   = 0;   /* not turning... */

 clear_count = 2;
 buf_num     = 0;

 icon_flag   = FALSE;
 gravity_flag = TRUE;

 landed      = FALSE;
 landings    = 0;
 score       = 0;
 fuel        = INIT_FUEL;
}


/*
 * 'init_gfx()' - open the window needed...
 */

void init_gfx(void)
{
 int x, y;
 Coord v[3];


 iconsize(84,66);
 minsize(384, 256);

 win = winopen("Lander 1.0 by Mike Sweet");

 doublebuffer();
 gconfig();

 qdevice(LEFTMOUSE);
 qdevice(MIDDLEMOUSE);
 qdevice(RIGHTMOUSE);
 qdevice(LEFTARROWKEY);
 qdevice(RIGHTARROWKEY);
 qdevice(QKEY);
 qdevice(GKEY);
 qdevice(PKEY);
 qdevice(ZEROKEY);
 qdevice(HOMEKEY);
 qdevice(WINFREEZE);
 qdevice(WINTHAW);

 mmode(MVIEWING);

 terrain_object = genobj();

 makeobj(terrain_object);

 for (x = 0; x < TERRAIN_SIZE; x ++)
  {
   bgnline();

   for (y = 0; y < TERRAIN_SIZE; y ++)
    {
     v[0] = (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v[2] = (y - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     if ((v[1] = terrain[x][y]) == 0.0 &&
         x > 0 && y > 0)
      {
       if (terrain[x - 1][y] == 0.0 &&
           terrain[x][y - 1] == 0.0 &&
           terrain[x - 1][y - 1] == 0.0)
        color(BLUE);
       else
        color(WHITE);
      }
     else
      color(WHITE);

     v3f(v);
    };

   endline();
  };

 for (y = 0; y < TERRAIN_SIZE; y ++)
  {
   bgnline();

   for (x = 0; x < TERRAIN_SIZE; x ++)
    {
     v[0] = (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v[2] = (y - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     if ((v[1] = terrain[x][y]) == 0.0 &&
         x > 0 && y > 0)
      {
       if (terrain[x - 1][y] == 0.0 &&
           terrain[x][y - 1] == 0.0 &&
           terrain[x - 1][y - 1] == 0.0)
        color(BLUE);
       else
        color(WHITE);
      }
     else
      color(WHITE);

     v3f(v);
    };

   endline();
  };

 closeobj();

 clear_object = genobj();

 makeobj(clear_object);

 color(BLACK);

 for (x = 0; x < TERRAIN_SIZE; x ++)
  {
   bgnline();

   for (y = 0; y < TERRAIN_SIZE; y ++)
    {
     v[0] = (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v[1] = terrain[x][y];
     v[2] = (y - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v3f(v);
    };

   endline();
  };

 for (y = 0; y < TERRAIN_SIZE; y ++)
  {
   bgnline();

   for (x = 0; x < TERRAIN_SIZE; x ++)
    {
     v[0] = (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v[1] = terrain[x][y];
     v[2] = (y - TERRAIN_SIZE / 2) * TERRAIN_SPACING;
     v3f(v);
    };

   endline();
  };

 closeobj();

/*
 * Define the mouse cursor...
 */

 curstype(C16X1);
 defcursor(1, cursor);
 curorigin(1, 8, 8);
 setcursor(1, 0, 0);

/*
 * Initialize things so they are drawn correctly...
 */

 reset_window();
}


/*
 * 'draw_window()' - redraw the contents of the window...
 */

void draw_window(void)
{
 char s[81];
 Icoord v[2];


/*
 * First, clear the window to BLACK...
 */

 mmode(MVIEWING);
 perspective(450, (double)xsize / (double)ysize, 0.1, 5000000.0);

 if (clear_count > 0)
  {
   color(BLACK);
   clear();

   clear_count --;
  };

/*
 * Next, save the current viewing matrix and rotate/translate as
 * necessary to orient the view...
 */

 pushmatrix();
  rotate(deltar[buf_num], 'z');
  rotate(rotation[buf_num] + view[buf_num], 'y');
  translate(-position[0][buf_num], -10.0 - position[1][buf_num],
            -position[2][buf_num]);
  callobj(clear_object);
 popmatrix();

 pushmatrix();
  rotate(deltar[2], 'z');
  rotate(rotation[2] + view[2], 'y');
  translate(-position[0][2], -10.0 - position[1][2], -position[2][2]);
  callobj(terrain_object);
 popmatrix();

 position[0][buf_num] = position[0][2];
 position[1][buf_num] = position[1][2];
 position[2][buf_num] = position[2][2];

 rotation[buf_num]    = rotation[2];
 view[buf_num]        = view[2];
 deltar[buf_num]      = deltar[2];

 buf_num ^= 1;

/*
 * Then, update the display in the status display...
 */

 if (!icon_flag)
  {
   ortho2(0, xsize - 1, 0, ysize - 1);

   color(BLACK);
   rectfi(0, ysize - 128, 400, ysize - 1);

   color(RED);

   sprintf(s, "Position:      (%6.0f, %6.0f, %6.0f)", position[0][2], position[1][2],
           position[2][2]);
   cmov2i(16, ysize - 32);
   charstr(s);

   sprintf(s, "Heading/View:  %-3d/%-3d", rotation[2] / 10, view[2] / 10);
   cmov2i(16, ysize - 64);
   charstr(s);

   sprintf(s, "Thrust/Motion: %-4.0f/(%-5.1f, %-5.1f, %-5.1f)", thrust,
           deltas[0], deltas[1] - GRAVITY, deltas[2]);
   cmov2i(16, ysize - 96);
   charstr(s);

   sprintf(s, "Fuel/Score:     %-6d/%-6d(%d)", fuel, score, landings);
   cmov2i(16, ysize - 128);
   charstr(s);

   if (fuel == 0)
    {
     cmov2i((xsize - 72) / 2, (ysize - 16) / 2);
     charstr("Game Over");
    };

   color(YELLOW);

   bgnline();
    v[0] = xsize / 2;
    v[1] = ysize / 2 - 10;
    v2i(v);
    v[1] += 20;
    v2i(v);
   endline();

   bgnline();
    v[0] = xsize / 2 - 10;
    v[1] = ysize / 2;
    v2i(v);
    v[0] += 20;
    v2i(v);
   endline();
  };
}


/*
 * 'move_ship()' - move the ship's current position...
 */

void move_ship(void)
{
 double radians;
 int x, z;
 double height;
 double temp;


/*
 * First, compute the angle of rotation in radians...
 */

 if ((rotation[2] += deltar[2]) < 0)
  rotation[2] += 3600;
 else if (rotation[2] >= 3600)
  rotation[2] -= 3600;

 radians = M_PI * rotation[2] / 1800.0;

/*
 * Add gravity constant...
 */

 if (!icon_flag && gravity_flag)
  deltas[1] -= GRAVITY;

/*
 * Then, move the ship forward in the direction given at 'thrust'
 * and up/down at 'deltah'...
 */

 if (fuel > fabs(thrust))
  fuel -= fabs(thrust);
 else
  {
   fuel = 0;
   thrust = 0;
  };

 if (fuel > FUEL_MASS)
  temp = thrust / (double)fuel * FUEL_MASS;
 else
  temp = thrust;

 deltas[0] += sin(radians) * temp;
 deltas[2] -= cos(radians) * temp;

 position[0][2] += deltas[0];
 position[1][2] += deltas[1];
 position[2][2] += deltas[2];

 x = position[0][2] / TERRAIN_SPACING + TERRAIN_SIZE / 2;
 z = position[2][2] / TERRAIN_SPACING + TERRAIN_SIZE / 2;

 if (x < 0 || z < 0 || x >= TERRAIN_SIZE || z >= TERRAIN_SIZE)
  {
   if (position[1][2] < 0.0)
    {
     position[1][2] = 0.0;
     deltas[1] = 0.0;
    };
  }
 else
  {
   height = (terrain[x][z] * (position[0][2] - (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING * (position[2][2] - (z - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING +

             terrain[x + 1][z] *
             (TERRAIN_SPACING - position[0][2] + (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING * (position[2][2] - (z - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING +

             terrain[x][z + 1] * (position[0][2] - (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING *
             (TERRAIN_SPACING - position[2][2] + (z - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING +

             terrain[x + 1][z + 1] *
             (TERRAIN_SPACING - position[0][2] + (x - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING *
             (TERRAIN_SPACING - position[2][2] + (z - TERRAIN_SIZE / 2) * TERRAIN_SPACING) /
             TERRAIN_SPACING) / 4;

   if (position[1][2] <= height && !landed)
    {
     position[1][2] = height;

     if (fabs(deltas[0]) > MAX_THRUST ||
         fabs(deltas[2]) > MAX_THRUST ||
         deltas[1] < (-2.0 * GRAVITY))
      {
       ringbell();
       score -= 10 * (landings / 5 + 1);
       fuel -= 100;
      }
     else
      {
       ringbell();
       landings ++;
       score += 50 * (landings / 5 + 1);
       fuel += 200;
       if (fuel > 4000)
        fuel = 4000;
      };

     deltas[0] = 0.0;
     deltas[1] = 0.0;
     deltas[2] = 0.0;

     landed = TRUE;
    }
   else if (position[1][2] > (height + 50))
    landed = FALSE;
   else if (position[1][2] < height)
    {
     position[1][2] = height;
     deltas[1] = 0.0;
    };
  };

 if (position[1][2] < 0.0)
  {
   position[1][2] = 0.0;
   deltas[0] = 0.0;
   deltas[1] = 0.0;
   deltas[2] = 0.0;
  };
}


/*
 * 'main()' - do everything...
 */

void main(argc, argv)
int argc;
char *argv[];
{
 short val;
 float temp;


/*
 * Initialize things...
 */

 fputs("Lander Version 1.0 by Mike Sweet.\n", stderr);

 initialize();

 if (argc > 1)
  if (strcmp(argv[1], "-f") == 0)
   prefposition(0, getgdesc(GD_XPMAX), 0, getgdesc(GD_YPMAX));

 init_gfx();

/*
 * Loop, showing terrain and handling events as necessary...
 */

 while (TRUE)
  {
   draw_window();
   swapbuffers();

   move_ship();

   if (getbutton(LEFTMOUSE) && !icon_flag && current_focus == win)
    {
     val = getvaluator(MOUSEX) - xorg;
     if (val >= 0 && val < xsize)
      deltar[2] = MAX_DELTAR * val / xsize - MAX_DELTAR / 2;

     val = getvaluator(MOUSEY) - yorg;
     if (val >= 0 && val < ysize && fuel > 0)
      {
       temp = (-2 * MAX_DELTAH * val / ysize + MAX_DELTAH);
       if ((fabs(temp) / 10) > fuel)
        fuel = 0;
       else
        fuel -= (fabs(temp) / 10);

       if (fuel > FUEL_MASS)
        temp = temp / (double)fuel * FUEL_MASS;

       deltas[1] += temp;
      };
    };
 
   while (qtest())
    switch (qread(&val))
     {
      case INPUTCHANGE   : current_focus = val;
                           break;

      case WINFREEZE     : icon_flag = TRUE;
                           reset_window();
                           break;

      case WINTHAW       : icon_flag = FALSE;
                           reset_window();
                           break;

      case REDRAW        :
      case REDRAWICONIC  : reset_window();
                           break;

      case MIDDLEMOUSE   : if (thrust > -MAX_THRUST && val &&
                               fuel > 0 && !icon_flag)
                            thrust -= MAX_THRUST * 0.1;

                           break;

      case LEFTMOUSE     : if (val == 0 && !icon_flag)
                            deltar[2] = 0;

                           break;

      case RIGHTMOUSE    : if (thrust < MAX_THRUST && val &&
                               fuel > 0 && !icon_flag)
                            thrust += MAX_THRUST * 0.1;

                           break;

      case LEFTARROWKEY  : if (val && !icon_flag)
                            view[2] -= 450;
                           if (view[2] < 0)
                            view[2] += 3600;

                           break;

      case RIGHTARROWKEY : if (val && !icon_flag)
                            view[2] += 450;
                           if (view[2] >= 3600)
                            view[2] -= 3600;

                           break;

      case HOMEKEY       : if (val && !icon_flag)
                            initialize();

                           break;

      case ZEROKEY       : if (val && !icon_flag)
                            thrust = 0.0;

                           break;

      case GKEY          : if (val && !icon_flag)
                            gravity_flag = !gravity_flag;

                           break;

      case PKEY          : if (val && !icon_flag)
                            {
                             draw_window();
                             cmov2i(16, 16);
                             color(RED);
                             charstr("PAUSED");
                             swapbuffers();
                             clear_count = 2;

                             while (TRUE)
                              if (qread(&val) == PKEY)
                               if (val)
                                break;
                            };

                           break;

      case QKEY          : if (icon_flag)
                            break;

                           winclose(winget());
                           exit(0);

                           break;
     };
  };
}




